import os

from .helper import Helper
from .email_ import Email

REQUIRED_VARS = [
    "EMAIL_USERNAME",
    "EMAIL_PASSWORD",
]

for var in REQUIRED_VARS:
    if not os.getenv(var):
        print(
            f"Environment variable {var} is missing, which is necessary for "
            "certain functions within this module."
        )



import csv
from pathlib import Path

from .fixtures import students_csv, parents_csv

with open(Path(Path(__file__).parent, "random_names.csv"), "r") as csvf:
    rd = csv.reader(csvf)
    names = [r for r in rd]


def test_make_students_csv(students_csv):
    # data after the header row
    data = students_csv[1:]

    assert data  # no empty list
    for row in data:
        # check that student names come from list of random names
        assert " ".join(row[:2]) in [" ".join(i) for i in names]

        # check that none of the teachers are also students
        teacher_last, teacher_first = row[3].split(", ")
        assert " ".join((teacher_first, teacher_last)) not in [
            " ".join(i) for i in names
        ]

    # check that homerooms are all in the same grade
    groups_by_teacher = {}
    for row in data:
        groups_by_teacher.setdefault(row[3], set())
        groups_by_teacher[row[3]].add(row[2])
    for set_ in groups_by_teacher.values():
        assert len(set_) == 1


def test_make_parents_csv(students_csv, parents_csv):
    students = students_csv[1:]
    parents = parents_csv[1:]

    assert parents  # no empty list
    for row in parents:
        # every row has 13 items
        assert len(row) == 13
        # row is all strings
        for i in row:
            assert isinstance(i, str)
        # unpack parent name variables
        parent_first_nm, parent_last_nm, *_ = row
        parent_name = " ".join((parent_first_nm, parent_last_nm))
        # no parent shares a name with a student
        assert parent_name not in [f"{s[0]} {s[1]}" for s in students]
        # no parent shares a name with a teacher
        teachers = set([" ".join(r[3].split(", ")[::-1]) for r in students])
        assert parent_name not in teachers
    # roughly two parents for each student
    assert -2 < len(parents) // 2 - len(students) < 2



class Group:
    """
    Group class is used to manage groups for extracurricular activities, field
    trips, and other purposes.
    """

    def __init__(self, name, grade_level, students):
        self.name = name
        self.grade_level = grade_level
        self.students = students


class Homeroom:
    def __init__(self, teacher, grade_level, students):
        """
        Ensure that string constants for csv headers of id, student names, and
        (if applicable) student emails are correct.
        """
        self.teacher = teacher
        self.grade_level = grade_level
        self.students = students


class Student:
    def __init__(self, context):
        context.setdefault("groups", [])
        context.setdefault("guardians", [])
        self.first_name = context.get("first_name")
        self.last_name = context.get("last_name")
        self.student_id = context.get("student_id")
        self.homeroom = context.get("homeroom")
        self.grade_level = context.get("grade_level")
        self.groups = context.get("groups")
        self.email = context.get("email")
        self.guardians = context.get("guardians")
        self.name = self.first_name + " " + self.last_name

        # primary_contact is an instance of ParentGuardian, which is assigned
        # during the parsing of parent / guardian data in the new_school_year
        # function within OnCourseMixin
        # TODO: assign attribute at init time, not later
        self.primary_contact = None

    def __eq__(self, other):
        if not isinstance(other, Student):
            return False

        if self.email is not None:
            return self.email == other.email

        if self.name is not None:
            return self.name == other.name

        raise ValueError("insufficient attributes to compare")

    def __str__(self, verbose=False):
        """
        Assemble nice printout of student information, and student's guardian
        information. Also, offer verbose option, which ties into /shell.py
        """
        data = [
            "--",
            self.name,
            self.grade_level,
            self.homeroom,
            self.email,
            "--" "\nGuardians:",
        ]
        student_data = [str(i) for i in data if i]  # filter out NoneTypes
        if not self.guardians:
            return "\n".join(data)
        # indent
        gu_data = ["\n\t".join(self.primary_contact.__str__().split("\n"))]
        cutoff = 3 if not verbose else len(self.guardians) - 1
        for gu in self.guardians[1:cutoff]:
            gu_data.append("\n\t".join(gu.__str__().split("\n")))
        return "\n".join(student_data + gu_data)

    def __repr__(self):
        normal = super().__repr__()
        return normal.replace("object", f'"{self.name}" object')


class ParentGuardian:
    def __init__(self, context: dict, verbose=False):

        self.student = context.get("student")
        self.first_name = context.get("first_name")
        self.last_name = context.get("last_name")
        self.home_phone = context.get("home_phone")
        self.mobile_phone = context.get("mobile_phone")
        self.work_phone = context.get("work_phone")
        self.email = context.get("email")
        self.relationship_to_student = context.get("relationship_to_student")
        self.primary_contact = context.get("primary_contact")
        self.allow_contact = context.get("allow_contact")
        self.student_resides_with = context.get("student_resides_with")

        # full name
        if self.first_name and self.last_name:
            self.name = self.first_name + " " + self.last_name
        else:
            raise ValueError(
                "First and last name were not provided as context to "
                "ParentGuardian class. This means that self.name cannot\n"
                "be constructed. Please at least pass a first and last name "
                "for every guardian into this class."
            )

        # warn about missing attributes
        for k, v in self.__dict__.items():
            if not v:
                if verbose:
                    print(f"WARNING: Guardian\t{self.name}\thas no value for\t{k}")

        # type checking
        if not isinstance(self.student, Student):
            raise ValueError("Student was a string, not a Student object.")
        for k, v in self.__dict__.items():
            if "phone" in k and v and not isinstance(v, int):
                raise ValueError("Phone number should be of type int.")

    def __str__(self):
        outs = [
            self.relationship_to_student,
            self.name,
            f"Contact allowed: {self.allow_contact}",
            f"Is primary contact: {self.primary_contact}",
            f"Mobile Phone: {self.mobile_phone}",
            f"Email Address: {self.email}",
            "\n",
        ]
        return "\n".join([str(i) for i in outs if i])



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from pathlib import Path
import shutil
import smtplib
import ssl
import os
from typing import Union

import markdown

from ._data_dir import get_data_dir


class Email:
    def __init__(self, username=None, password=None):
        self.email_addr = username or os.getenv("EMAIL_USERNAME", "")
        self.password = password or os.getenv("EMAIL_PASSWORD", "")
        self.connection = None

        # create ~/.teacherhelper/email_templates if needed
        self.template_dir = get_data_dir() / ".teacherhelper" / "email_templates"
        if not self.template_dir.exists():
            os.makedirs(self.template_dir)
        default_template = Path(self.template_dir, "default.html")
        if not default_template.exists():
            shutil.copyfile(
                Path(Path(__file__).parent, "email_default_template.html"),
                default_template,
            )

    def __enter__(self):
        context = ssl.create_default_context()
        self.connection = smtplib.SMTP_SSL(
            "smtp.gmail.com",
            port=465,
            context=context,
        )
        self.connection.login(
            self.email_addr,
            self.password,
        )
        return self

    def __exit__(self, *_):
        if self.connection:
            self.connection.close()
        self.connection = None

    def send(
        self,
        *,
        to: Union[str, list],
        subject: str,
        message: str,
        cc: Union[list, str] = None,
        bcc: Union[list, str] = None,
        template_name: str = "default.html"
    ):
        """
        Simple utility for sending an email. Helpful for mail merges!

        *message* should be a string of markdown text, which will be
        converted into html and plain text email attachments.

        *template_name* is the name of an html email template in
        ~/.teacherhelper/email_templates. An email template can be any html
        file with the template tag `{{ email_content }}` in it. The markdown
        input will be converted into html, and that html will replace the
        `{{ email_content }}` tag.
        """
        # ensure that to, cc, and bcc are strings
        to = to if isinstance(to, str) else ", ".join(to)
        cc = cc if isinstance(cc, str) else ", ".join(cc) if cc else ""
        bcc = bcc if isinstance(bcc, str) else ", ".join(bcc) if bcc else ""

        if not self.connection:
            raise ValueError(
                "Connection must be established in __enter__. Use this class "
                "as a context manager"
            )
        me = self.email_addr
        msg = MIMEMultipart("alternative")
        msg["Subject"] = subject
        msg["From"] = me
        msg["To"] = to
        if cc:
            msg["Cc"] = cc
        if bcc:
            msg["Bcc"] = bcc

        html = self.make_html_message(message, template_name)

        part1 = MIMEText(message, "plain")
        part2 = MIMEText(html, "html")
        msg.attach(part1)
        msg.attach(part2)
        self.connection.sendmail(me, to, msg.as_string())

    def make_html_message(
        self,
        markdown_message: str,
        template_name: str = "default.html",
    ):
        template_path = Path(self.template_dir, template_name)

        with open(template_path, "r") as fp:
            template = fp.read()
            html = template.replace(
                "{{ email_content }}", markdown.markdown(markdown_message)
            )

        return html



import re

import docx
from docx.opc.constants import RELATIONSHIP_TYPE as RT


def regex_search(st, doc, regex, near_match_regex=r""):
    """Iterates through all paragraphs, tables, and hyperlinks in a word doc,
    and extracts any strings that match the regex and, optionally, the near
    match regex. It returns a list of match tuplets and miss tuplets:
        matches = [
            (st.name, match.string, st.homeroom),
            ...
        ]
        misses = [
            (st.name, miss.string, st.homeroom),
            ...
        ]
    """
    matches = []
    misses = []
    document = docx.Document(doc)
    # search paragraphs
    for par in document.paragraphs:
        mo = re.fullmatch(regex, par.text.strip())
        if mo:
            matches.append((st.name, mo.string, st.homeroom))
        else:
            mo = re.search(near_match_regex, par.text)
            if mo:
                misses.append((st.name, mo.string, st.homeroom))
    # search tables
    for table in document.tables:
        for row in table.rows:
            for cell in row.cells:
                mo = re.fullmatch(regex, cell.text.strip())
                if mo:
                    matches.append((st.name, mo.string, st.homeroom))
                else:
                    mo = re.search(near_match_regex, cell.text)
                    if mo:
                        misses.append((st.name, mo.string, st.homeroom))
    # search for links
    for rel in document.part.rels:
        Rel = document.part.rels[rel]
        if Rel.reltype == RT.HYPERLINK:
            text = Rel._target
            mo = re.fullmatch(regex, cell.text.strip())
            if mo:
                matches.append((st.name, mo.string, st.homeroom))
            else:
                mo = re.search(near_match_regex, cell.text)
                if mo:
                    misses.append((st.name, mo.string, st.homeroom))
    if near_match_regex:
        return matches, misses
    return matches, None



import os
import logging
import dbm
import shelve
from typing import Union, cast, Dict
from datetime import datetime

from fuzzywuzzy import process

from ._data_dir import get_data_dir
from .entities import Group, Homeroom, ParentGuardian, Student
from .oncourse_mixin import OnCourseMixin


logger = logging.getLogger(__name__)


class Helper(OnCourseMixin):
    """Public API for the module, through which students and parents can be
    searched for.
    """

    def __init__(
        self,
        homerooms: Dict[str, Homeroom],
        students: Dict[str, Student],
        groups: Dict[str, Group],
    ):
        self.homerooms = homerooms
        self.students = students
        self.groups = groups
        self.cache_dir = os.path.join(__file__, "cache")

    def write_cache(self):
        with shelve.open(os.path.join(get_data_dir(), "cache"), "c") as db:
            db["data"] = self
            db["date"] = datetime.now()

    def find_nearest_match(
        self, student_name: str, threshold=90, **_
    ) -> Union[Student, None]:
        """
        Returns a student object. If auto_yes=True, it will presume that the
        best matching student is correct. Optionally, set a levenshtien distance
        threshold below which students will not be included.
        """
        if not isinstance(student_name, str):
            raise Exception("Student name must be a string")

        # direct match
        if st := self.students.get(student_name.title()):
            logger.debug(f"Exact match for {st.name}")
            return st

        # get nearest match
        if result := process.extractOne(student_name, self.students.keys()):
            closest_name, confidence = result[0], result[1]
            if confidence >= threshold:
                return self.students[closest_name]

    def find_parent(self, parent_name: str) -> Union[ParentGuardian, None]:
        """Return a parent matching the given name, preferentially searching
        amongst primary contacts. May return `None` if there is not a close
        match."""
        # build mapping of parent name to string so we can perform a fuzzy
        # search. Make separate collections of all_guardians and
        # primary_guardians so we can prefer to match primaries
        all_guardians = {}
        primary_guardians = {}
        for st in self.students.values():
            for g in st.guardians:
                all_guardians.setdefault(g.name, g)
                if g.primary_contact:
                    primary_guardians.setdefault(g.name, g)
            primary_contact = cast(ParentGuardian, st.primary_contact)
            if primary_contact is not None:
                primary_guardians.setdefault(primary_contact.name, primary_contact)
            # insert the other guardians only if there is no primary
            else:
                for g in st.guardians:
                    all_guardians.setdefault(g.name, g)

        # prefer match amongst primary contacts
        primary_match = process.extractOne(
            parent_name, [g.name for g in primary_guardians.values()]
        )
        if primary_match and primary_match[1] > 85:
            if mo := primary_guardians.get(primary_match[0]):
                return mo.student

        # search all parents and guardians otherwise
        name_match = process.extractOne(
            parent_name, [g.name for g in all_guardians.values()]
        )
        if name_match and (mo := all_guardians.get(name_match[0])):
            return mo.student

    @classmethod
    def read_cache(cls, check_date=True):
        """
        This static method returns a class because I like to break the rules.
        there's a reason for the rules; this garbage doesn't work
        """
        with shelve.open(os.path.join(get_data_dir(), "cache"), "r") as db:
            cls = cast(cls, db["data"])
            date: datetime = cast(datetime, db["date"])

        if check_date and (
            datetime.now().month in range(9, 12) and date.month in range(1, 7)
        ):
            raise Exception(
                "It appears that the cache is from last school year. Please\n"
                "provide new data, re-instantiate, and re-write cache."
            )
        return cls

    @staticmethod
    def cache_exists():
        try:
            sh = shelve.open(os.path.join(get_data_dir(), "cache"), "r")
            sh.close()
            return True
        except dbm.error:  # type: ignore
            return False


def get_helper() -> Helper:
    return Helper.read_cache()



"""
Script to generate mock data.
"""

import csv
from pathlib import Path
from itertools import cycle
from typing import List
import random

import pytest


with open(Path(Path(__file__).parent, "random_names.csv"), "r") as csvf:
    rd = csv.reader(csvf)
    names = [r for r in rd]


@pytest.fixture
def students_csv() -> list:
    teachers = [", ".join((n1[0], n2[1])) for n1, n2 in zip(names[:30], names[1:31])]
    teacher_grade = {teacher: f"{random.randint(4, 7)}th Grade" for teacher in teachers}
    rows = [
        [
            "First Name",
            "Last Name",
            "Grade Level",
            "Homeroom Teacher",
            "Email Address 1",
            "Birth Date",
        ]
    ]
    for teacher, name in zip(cycle(teachers), names[:300]):
        first, last = name
        # Each row is:
        # first, last, grade level, homeroom teacher, email, birthday
        rows.append(
            [
                first,
                last,
                teacher_grade[teacher],
                teacher,
                str(hash(first + last))[1:] + "@empacad.org",
                # birthday
                f"{random.randint(1, 13)}/{random.randint(1,28)}/{random.randint(2005, 2010)}",
            ]
        )
    return rows


@pytest.fixture
def parents_csv(students_csv):
    """
    Need to take the student data to make random parents for those students.

    Each row has:
    - Guardian First Name
    - Guardian Last Name
    - Student First Name
    - Student Last Name
    - Primary Contact
    - Guardian Email Address 1
    - Guardian Mobile Phone
    - Guardian Phone
    - Guardian Work Phone
    - Comments
    - Allow Contact
    - Student Resides With
    - Relation to Student
    """
    students_csv_data = students_csv[1:]  # skip header row
    parent_names = names[300:900]
    rows = [
        [
            "Guardian First Name",
            "Guardian Last Name",
            "Student First Name",
            "Student Last Name",
            "Primary Contact",
            "Guardian Email Address 1",
            "Guardian Mobile Phone",
            "Guardian Phone",
            "Guardian Work Phone",
            "Comments",
            "Allow Contact",
            "Student Resides With",
            "Relation to Student",
        ]
    ]
    for parent_name, student_row in zip(parent_names, cycle(students_csv_data)):
        rows.append(
            [
                *parent_name,  # guardian full name
                student_row[0],  # student first name
                student_row[1],  # student last name
                "Y",  # is primary contact
                f"{str(hash(''.join(parent_name)))[1:]}@gmail.com",  # email
                str(random.randint(9730000000, 9739999999)),  # mobile phone
                *[""] * 2,  # home & work phone
                str(hash(tuple(student_row))),  # comments
                ["Y", "N"][random.randint(0, 1)],  # allow contact
                ["Y", "N"][random.randint(0, 1)],  # student resides with
                ["Mother", "Father", "Other", "Grandmother", "Grandfather"][
                    random.randint(0, 4)
                ],  # relation to student
            ]
        )
    return rows



import csv
from pathlib import Path

from ._data_dir import get_data_dir
from .tools.csv_parser import IterCsv
from .entities import Homeroom, ParentGuardian, Student


class OnCourseMixin:
    """This doesn't specifically interface with the OnCourse API, it just
    initializes the helper cache with the spreadsheet reports that I output
    from oncourse, using those specific headers"""

    @classmethod
    def new_school_year(cls):
        """Take a spreadsheet of student data and parent data, build the
        necessary relationships.

        Student data csv should include columns:

        - first name
        - last name
        - grade level
        - homeroom teacher
        - email address 1
        - birth date

        Parent data csv should include:

        - guardian first name
        - guardian last name
        - student first name
        - student last name
        - primary contact
        - guardian email address 1
        - guardian mobile phone
        - guardian phone
        - guardian work phone
        - comments
        - allow contact
        - student resides with
        - relation to student
        """
        student_data = Path(get_data_dir(), "students.csv")
        guardian_data = Path(get_data_dir(), "parents.csv")
        STUDENTS = {}
        HOMEROOMS = {}

        with open(student_data, "r", encoding="utf-8-sig") as csvfile:
            rows = [r for r in csv.reader(csvfile, delimiter=",")]
        acceptable_headers = [
            "first name",
            "last name",
            "grade level",
            "homeroom teacher",
            "email address 1",
            "birth date",
        ]
        for get_item in IterCsv(acceptable_headers, rows):
            (first, last, grade, homeroom, email, birthday,) = (
                get_item("first name"),
                get_item("last name"),
                get_item("grade level"),
                get_item("homeroom teacher"),
                get_item("email address 1"),
                get_item("birth date"),
            )
            # convert grade to int
            for i in [4, 5, 6, 7]:
                if str(i) in grade:
                    grade = i
                    break
            student = Student(
                {
                    "first_name": first,
                    "last_name": last,
                    "grade_level": grade,
                    "homeroom": homeroom,
                    "email": email,
                    "birthday": birthday,
                }
            )
            STUDENTS[first + " " + last] = student
            if homeroom not in HOMEROOMS:
                HOMEROOMS[homeroom] = Homeroom(
                    homeroom,
                    grade,
                    [student],
                )
            else:
                HOMEROOMS[homeroom].students.append(student)

        # instantiation
        self = cls(HOMEROOMS, STUDENTS, {})  # type: ignore
        with open(guardian_data, "r", encoding="utf8") as csvfile:
            rows = [r for r in csv.reader(csvfile)]
        acceptable_headers = [
            "guardian first name",
            "guardian last name",
            "student first name",
            "student last name",
            "primary contact",
            "guardian email address 1",
            "guardian mobile phone",
            "guardian phone",
            "guardian work phone",
            "comments",
            "allow contact",
            "student resides with",
            "relation to student",
        ]
        for get_item in IterCsv(acceptable_headers, rows):
            raw_context = {
                "first_name": get_item("guardian first name"),
                "last_name": get_item("guardian last name"),
                "student": get_item("student first name")
                + " "
                + get_item("student last name"),
                "primary_contact": get_item("primary contact"),
                "email": get_item("guardian email address 1"),
                "mobile_phone": get_item("guardian mobile phone"),
                "home_phone": get_item("guardian phone"),
                "work_phone": get_item("guardian work phone"),
                "comments": get_item("comments"),
                "allow_contact": get_item("allow contact"),
                "student_resides_with": get_item("student resides with"),
                "relationship_to_student": get_item("relation to student"),
            }
            clean_context = {}
            # find student object match
            student = self.find_nearest_match(  # type: ignore
                raw_context["student"],
            )
            if not student:
                continue
            if student.name != raw_context["student"]:
                raise ValueError(
                    f"Integrity error. {student.name} does not equal "
                    + raw_context["student"]
                )
            clean_context["student"] = student
            for k, v in raw_context.items():
                # clean phone numbers
                if "phone" in k:
                    if "_" in v:
                        continue
                    nums = [l for l in v if l.isnumeric()]
                    if not nums:
                        continue
                    if len(nums) < 10:
                        continue
                    if len(nums) > 11:
                        raw_context["comments"] += f"\n{k} is {v}"
                        continue
                    try:
                        phone_number = int("".join(nums))
                        clean_context[k] = phone_number
                    except TypeError:
                        continue
                # convert boolean fields to boolean
                if k in [
                    "primary_contact",
                    "allow_contact",
                    "student_resides_with",
                ]:
                    if "Y" in v:
                        v = True
                    elif "N" in v:
                        v = False
                    else:
                        raise ValueError(
                            f"Supposedly boolean field {k} could not"
                            "be converted into a boolean value."
                        )
                    clean_context[k] = v
                clean_context.setdefault(k, v)
            parent = ParentGuardian(clean_context)
            student.guardians.append(parent)
            if clean_context["primary_contact"]:
                # It is important that the primary_contact attribute of the
                # student is assigned while parent / guardian data is being
                # parsed.
                student.primary_contact = parent
        return self
